package at.jku.students.textsearch;

import at.jku.ssw.Iterator;
import at.jku.ssw.list.List;
import at.jku.ssw.textsearch.Alice;
import at.jku.ssw.textsearch.TextSearcher;
import at.jku.ssw.util.Out;

class TextSearchMain {

	public static void main(String[] args) {
		testSimple();
		markMatches();
		benchmark();
	}

	public static void testSimple() {
		test("\n*** Test BruteForce ***", new BruteForceSearcher());
		test("\n*** Test BoyerMoore ***", new BoyerMooreSearcher());
	}

	/**
	 * 
	 */
	private static void test(String title, TextSearcher searcher) {
		Out.println(title);
		countAndPrint(searcher, "wettergott", "Gott", false, false);
		countAndPrint(searcher, "wettergott", "Gott", true, false);
		countAndPrint(searcher, "wettergott", "tt", false, false);
		countAndPrint(searcher, "wettergott", "t", false, false);
		countAndPrint(searcher, "Er ist Kaiser", "er", false, false);
		countAndPrint(searcher, "Er ist Kaiser", "er", false, true);
		countAndPrint(searcher, "Er ist Kaiser", "kaiser", false, true);
		countAndPrint(searcher, "Er ist Kaiser", "er", true, false);
		countAndPrint(searcher, "Er ist Kaiser", "er", true, true);
		countAndPrint(searcher, "Er ist Kaiser", "Kaiser", true, true);
		Out.println();
		countAndPrint(searcher, "Annanna", "Anna", false, false);
		countAndPrint(searcher, "Annanna", "Anna", true, false);
		countAndPrint(searcher, "Annanna", "Annanna", false, true);
		Out.println();

	}

	private static void countAndPrint(TextSearcher search, String text, String pattern, boolean matchCase,
			boolean wholeWordOnly) {
		Out.println(String.format("count matches for (\"%s\", \"%s\", %b, %b)=%d", text, pattern, matchCase,
				wholeWordOnly, search.findMatches(text, pattern, matchCase, wholeWordOnly).size()));
	}

	/**
	 * 
	 */
	public static void markMatches() {
		printOccurrences("Whiskeymixer", new BruteForceSearcher(),
				"Der Whiskeymixer mixt frischen Whiskey.\nFrischen Whiskey mixt der Whiskeymixer", "Whiskey", true,
				false);
		printOccurrences("Alice", new BoyerMooreSearcher(), Alice.getText(), Alice.getPatterns()[0], false, true);
	}

	private static void printOccurrences(String title, TextSearcher searcher, String text, String pattern,
			boolean matchCase, boolean wholeWordOnly) {
		Out.println(String.format("*** %s: Mark all matches of '%s' ***", title, pattern));
		Out.println();
		List matches = searcher.findMatches(text, pattern, matchCase, wholeWordOnly);

		Iterator it = matches.iterator();
		if (it.hasNext()) {
			int nextMatch = (int) it.next();
			int startIdx = 0;

			for (String line : text.split(System.lineSeparator())) {
				int endIdx = startIdx + line.length() + System.lineSeparator().length();
				// print original line
				Out.println(line);

				StringBuffer sb = new StringBuffer();
				int markerCount = startIdx;
				while (nextMatch < endIdx) {
					int length = nextMatch - markerCount;
					appendSpaces(sb, length);
					sb.append('^');
					markerCount = nextMatch + 1;
					nextMatch = it.hasNext() ? (int) it.next() : Integer.MAX_VALUE;
				}
				// print marker line
				Out.println(sb.toString());
				startIdx = endIdx;
			}
		}
		Out.println();
		Out.println(String.format("Occurrences of '%s' (matchCase=%b wholeWord=%b): %d", pattern, matchCase,
				wholeWordOnly, matches.size()));
		Out.println("---------------------");
		Out.println();

	}

	private static void appendSpaces(StringBuffer sb, int length) {
		for (int i = 0; i < length; i++) {
			sb.append(' ');
		}
	}

	/**
	 * 
	 */
	public static void benchmark() {
		Out.println("\n*** Benchmark ***");
		evaluate(Alice.getText().toLowerCase(), Alice.getPatterns());
	}

	private static class Result {

		int cnt = 0;
		long start = 0;
		long end = 0;
		long time = 0;

		void add(Result r) {
			this.cnt += r.cnt;
			this.time += r.time;
		}
	}

	private static void evaluate(String text, String[] patterns) {

		TextSearcher brute = new BruteForceSearcher();
		TextSearcher boyer = new BoyerMooreSearcher();
		Result total1 = new Result(), total2 = new Result();
		String header = String.format("\n%-20s %4s %6s ms   %4s %6s ms", "Pattern", "#", "Brute", "#", "Boyer");
		Out.println(header);
		Out.println("---------------------------------------------");
		for (String p : patterns) {
			Result r1 = evaluate(brute, text, p);
			Result r2 = evaluate(boyer, text, p);
			Out.println(String.format("%-20s %4d %6d ms   %4d %6d ms", p, r1.cnt, r1.time, r2.cnt, r2.time));
			total1.add(r1);
			total2.add(r2);
			assert r1.cnt == r2.cnt;
		}
		Out.println();
		Out.println(header);
		Out.println("---------------------------------------------");
		Out.println(String.format("%25d %6d ms   %4d %6d ms", total1.cnt, total1.time, total2.cnt, total2.time));
	}

	private static Result evaluate(TextSearcher searcher, String text, String pattern) {
		final int REPEAT = 100;
		Result result = new Result();
		result.start = System.currentTimeMillis();
		for (int i = 0; i < REPEAT; i++) {
			result.cnt = searcher.findMatches(text, pattern, false, true).size();
		}
		result.end = System.currentTimeMillis();
		result.time = (result.end - result.start);
		return result;
	}
}
